@@ -38,21 +38,7 @@
 #include <stdlib.h>
 
 #ifdef G_OS_WIN32
-#include <io.h>
-#ifndef F_OK
-#define	F_OK 0
-#define	W_OK 2
-#define	R_OK 4
-#endif /* !F_OK */
-
-#ifndef S_ISREG
-#define S_ISREG(mode) ((mode)&_S_IFREG)
-#endif
-
-#ifndef S_ISDIR
-#define S_ISDIR(mode) ((mode)&_S_IFDIR)
-#endif
-
+#include <windows.h>
 #endif /* G_OS_WIN32 */
 
 #ifndef S_ISLNK
@@ -68,7 +54,7 @@
 
 /**
  * g_file_test:
- * @filename: a filename to test
+ * @filename: a filename to test in the GLib file name encoding
  * @test: bitfield of #GFileTest flags
  * 
  * Returns %TRUE if any of the tests in the bitfield @test are
@@ -86,7 +72,7 @@
  * %TRUE for %G_FILE_TEST_IS_SYMLINK and %FALSE for all other flags.
  *
  * You should never use g_file_test() to test whether it is safe
- * to perform an operaton, because there is always the possibility
+ * to perform an operation, because there is always the possibility
  * of the condition changing before you actually perform the operation.
  * For example, you might think you could use %G_FILE_TEST_IS_SYMLINK
  * to know whether it is is safe to write to a file without being
@@ -104,9 +90,15 @@
  * %G_FILE_TEST_IS_EXECUTABLE are implemented using the access()
  * system call. This usually doesn't matter, but if your program
  * is setuid or setgid it means that these tests will give you
- * the answer for the real user ID and group ID , rather than the
+ * the answer for the real user ID and group ID, rather than the
  * effective user ID and group ID.
  *
+ * On Windows, there are no symlinks, so testing for
+ * %G_FILE_TEST_IS_SYMLINK will always return %FALSE. Testing for
+ * %G_FILE_TEST_IS_EXECUTABLE will just check that the file exists and
+ * its name indicates that it is executable, checking for well-known
+ * extensions and those listed in the %PATHEXT environment variable.
+ *
  * Return value: whether a test was %TRUE
  **/
 gboolean
@@ -114,6 +106,8 @@ g_file_test (const gchar *filename,
              GFileTest    test)
 {
 #ifdef G_OS_WIN32
+  int attributes;
+
   if (G_WIN32_HAVE_WIDECHAR_API ())
     {
       wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
@@ -121,91 +115,85 @@ g_file_test (const gchar *filename,
       if (wfilename == NULL)
 	return FALSE;
 
-      if ((test & G_FILE_TEST_EXISTS) && (_waccess (wfilename, F_OK) == 0))
-	{
-	  g_free (wfilename);
-	  return TRUE;
-	}
-      
-      if (test & (G_FILE_TEST_IS_REGULAR |
-		  G_FILE_TEST_IS_DIR |
-		  G_FILE_TEST_IS_EXECUTABLE))
-	{
-	  struct _stat s;
-	  
-	  if (_wstat (wfilename, &s) == 0)
-	    {
-	      if ((test & G_FILE_TEST_IS_REGULAR) && S_ISREG (s.st_mode))
-		{
-		  g_free (wfilename);
-		  return TRUE;
-		}
-	      
-	      if ((test & G_FILE_TEST_IS_DIR) && S_ISDIR (s.st_mode))
-		{
-		  g_free (wfilename);
-		  return TRUE;
-		}
-	      
-	      if ((test & G_FILE_TEST_IS_EXECUTABLE) &&
-		  (s.st_mode & _S_IEXEC))
-		{
-		  g_free (wfilename);
-		  return TRUE;
-		}
-	    }
-	}
+      attributes = GetFileAttributesW (wfilename);
 
       g_free (wfilename);
-      
-      return FALSE;
     }
   else
     {
-      gchar *cp_filename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
+      gchar *cpfilename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
 
-      if (cp_filename == NULL)
+      if (cpfilename == NULL)
 	return FALSE;
       
-      if ((test & G_FILE_TEST_EXISTS) && (access (cp_filename, F_OK) == 0))
-	{
-	  g_free (cp_filename);
-	  return TRUE;
-	}
+      attributes = GetFileAttributesA (cpfilename);
       
-      if (test & (G_FILE_TEST_IS_REGULAR |
-		  G_FILE_TEST_IS_DIR |
-		  G_FILE_TEST_IS_EXECUTABLE))
+      g_free (cpfilename);
+    }
+
+  if (attributes == INVALID_FILE_ATTRIBUTES)
+    return FALSE;
+
+  if (test & G_FILE_TEST_EXISTS)
+    return TRUE;
+      
+  if (test & G_FILE_TEST_IS_REGULAR)
+    return (attributes & (FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_DEVICE)) == 0;
+
+  if (test & G_FILE_TEST_IS_DIR)
+    return (attributes & FILE_ATTRIBUTE_DIRECTORY) != 0;
+
+  if (test & G_FILE_TEST_IS_EXECUTABLE)
+    {
+      const gchar *lastdot = strrchr (filename, '.');
+      const gchar *pathext = NULL, *p;
+      int extlen;
+
+      if (lastdot == NULL)
+	return FALSE;
+
+      if (stricmp (lastdot, ".exe") == 0 ||
+	  stricmp (lastdot, ".cmd") == 0 ||
+	  stricmp (lastdot, ".bat") == 0 ||
+	  stricmp (lastdot, ".com") == 0)
+	return TRUE;
+
+      /* Check if it is one of the types listed in %PATHEXT% */
+
+      pathext = g_getenv ("PATHEXT");
+      if (pathext == NULL)
+	return FALSE;
+
+      pathext = g_utf8_casefold (pathext, -1);
+
+      lastdot = g_utf8_casefold (lastdot, -1);
+      extlen = strlen (lastdot);
+
+      p = pathext;
+      while (TRUE)
 	{
-	  struct stat s;
-	  
-	  if (stat (cp_filename, &s) == 0)
+	  const gchar *q = strchr (p, ';');
+	  if (q == NULL)
+	    q = p + strlen (p);
+	  if (extlen == q - p &&
+	      memcmp (lastdot, p, extlen) == 0)
 	    {
-	      if ((test & G_FILE_TEST_IS_REGULAR) && S_ISREG (s.st_mode))
-		{
-		  g_free (cp_filename);
-		  return TRUE;
-		}
-	      
-	      if ((test & G_FILE_TEST_IS_DIR) && S_ISDIR (s.st_mode))
-		{
-		  g_free (cp_filename);
-		  return TRUE;
-		}
-	      
-	      if ((test & G_FILE_TEST_IS_EXECUTABLE) &&
-		  (s.st_mode & _S_IEXEC))
-		{
-		  g_free (cp_filename);
-		  return TRUE;
-		}
+	      g_free ((gchar *) pathext);
+	      g_free ((gchar *) lastdot);
+	      return TRUE;
 	    }
+	  if (*q)
+	    p = q + 1;
+	  else
+	    break;
 	}
 
-      g_free (cp_filename);
-      
+      g_free ((gchar *) pathext);
+      g_free ((gchar *) lastdot);
       return FALSE;
     }
+
+  return FALSE;
 #else
   if ((test & G_FILE_TEST_EXISTS) && (access (filename, F_OK) == 0))
     return TRUE;
@@ -741,18 +729,21 @@ get_contents_win32 (const gchar *filename,
  * g_file_get_contents:
  * @filename: name of a file to read contents from, in the GLib file name encoding
  * @contents: location to store an allocated string
- * @length: location to store length in bytes of the contents
- * @error: return location for a #GError
+ * @length: location to store length in bytes of the contents, or %NULL
+ * @error: return location for a #GError, or %NULL
  * 
  * Reads an entire file into allocated memory, with good error
- * checking. If @error is set, %FALSE is returned, and @contents is set
- * to %NULL. If %TRUE is returned, @error will not be set, and @contents
- * will be set to the file contents.  The string stored in @contents
- * will be nul-terminated, so for text files you can pass %NULL for the
- * @length argument.  The error domain is #G_FILE_ERROR. Possible
- * error codes are those in the #GFileError enumeration.
+ * checking. 
  *
- * Return value: %TRUE on success, %FALSE if error is set
+ * If the call was successful, it returns %TRUE and sets @contents to the file 
+ * contents and @length to the length of the file contents in bytes. The string 
+ * stored in @contents will be nul-terminated, so for text files you can pass 
+ * %NULL for the @length argument. If the call was not successful, it returns 
+ * %FALSE and sets @error. The error domain is #G_FILE_ERROR. Possible error  
+ * codes are those in the #GFileError enumeration. In the error case, 
+ * @contents is set to %NULL and @length is set to zero.
+ *
+ * Return value: %TRUE on success, %FALSE if an error occurred
  **/
 gboolean
 g_file_get_contents (const gchar *filename,
