@@ -39,8 +39,13 @@
 #endif
 #ifdef G_OS_WIN32
 #include <io.h>
+
+#ifndef S_ISREG
+#define S_ISREG(mode) ((mode)&_S_IFREG)
 #endif
 
+#endif  /* G_OS_WIN23 */
+
 #include "gconvert.h"
 #include "gdataset.h"
 #include "gerror.h"
@@ -359,7 +364,7 @@ g_key_file_load_from_fd (GKeyFile       *key_file,
       return FALSE;
     }
 
-  if ((stat_buf.st_mode & S_IFMT) != S_IFREG)
+  if (!S_ISREG (stat_buf.st_mode))
     {
       g_set_error (error, G_KEY_FILE_ERROR,
                    G_KEY_FILE_ERROR_PARSE,
@@ -3036,7 +3041,7 @@ g_key_file_parse_value_as_string (GKeyFile     *key_file,
       p++;
     }
 
-  if (p[-1] == '\\' && error == NULL)
+  if (p > value && p[-1] == '\\' && error == NULL)
     g_set_error (error, G_KEY_FILE_ERROR,
 		 G_KEY_FILE_ERROR_INVALID_VALUE,
 		 _("Key file contains escape character at end of line"));
