@@ -372,6 +372,9 @@ print_entry (GOptionGroup       *group,
   if (entry->flags & G_OPTION_FLAG_HIDDEN)
     return;
 
+  if (entry->long_name[0] == 0)
+    return;
+
   str = g_string_new (NULL);
   
   if (entry->short_name)
@@ -1057,8 +1060,10 @@ free_pending_nulls (GOptionContext *context,
  *
  * If the parsing is successful, any parsed arguments are
  * removed from the array and @argc and @argv are updated 
- * accordingly. In case of an error, @argc and @argv are
- * left unmodified.
+ * accordingly. A '--' option is stripped from @argv
+ * unless there are unparsed options before and after it, 
+ * or some of the options after it start with '-'. In case 
+ * of an error, @argc and @argv are left unmodified. 
  *
  * If automatic <option>--help</option> support is enabled
  * (see g_option_context_set_help_enabled()), and the 
@@ -1120,6 +1125,8 @@ g_option_context_parse (GOptionContext   *context,
   if (argc && argv)
     {
       gboolean stop_parsing = FALSE;
+      gboolean has_unknown = FALSE;
+      gint separator_pos = 0;
 
       for (i = 1; i < *argc; i++)
 	{
@@ -1137,7 +1144,7 @@ g_option_context_parse (GOptionContext   *context,
 		  /* '--' terminates list of arguments */
 		  if (*arg == 0)
 		    {
-		      add_pending_null (context, &((*argv)[i]), NULL);
+		      separator_pos = i;
 		      stop_parsing = TRUE;
 		      continue;
 		    }
@@ -1306,11 +1313,14 @@ g_option_context_parse (GOptionContext   *context,
 		    }
 		}
 	      
+	      if (!parsed)
+		has_unknown = TRUE;
+
 	      if (!parsed && !context->ignore_unknown)
 		{
 		  g_set_error (error,
 			       G_OPTION_ERROR, G_OPTION_ERROR_UNKNOWN_OPTION,
-			       _("Unknown option %s"), (*argv)[i]);
+				   _("Unknown option %s"), (*argv)[i]);
 		  goto fail;
 		}
 	    }
@@ -1322,32 +1332,41 @@ g_option_context_parse (GOptionContext   *context,
 					argc, argv, error, &parsed))
 		goto fail;
 	      
+	      if (!parsed && (has_unknown || (*argv)[i][0] == '-'))
+		separator_pos = 0;
 	    }
 	}
 
-      /* Call post-parse hooks */
-      list = context->groups;
-      while (list)
-	{
-	  GOptionGroup *group = list->data;
+      if (separator_pos > 0)
+	add_pending_null (context, &((*argv)[separator_pos]), NULL);
+	
+    }
 
-	  if (group->post_parse_func)
-	    {
-	      if (!(* group->post_parse_func) (context, group,
-					       group->user_data, error))
-		goto fail;
-	    }
-	  
-	  list = list->next;
-	}
-
-      if (context->main_group && context->main_group->post_parse_func)
+  /* Call post-parse hooks */
+  list = context->groups;
+  while (list)
+    {
+      GOptionGroup *group = list->data;
+      
+      if (group->post_parse_func)
 	{
-	  if (!(* context->main_group->post_parse_func) (context, context->main_group,
-							 context->main_group->user_data, error))
+	  if (!(* group->post_parse_func) (context, group,
+					   group->user_data, error))
 	    goto fail;
 	}
-
+      
+      list = list->next;
+    }
+  
+  if (context->main_group && context->main_group->post_parse_func)
+    {
+      if (!(* context->main_group->post_parse_func) (context, context->main_group,
+						     context->main_group->user_data, error))
+	goto fail;
+    }
+  
+  if (argc && argv)
+    {
       free_pending_nulls (context, TRUE);
       
       for (i = 1; i < *argc; i++)
