@@ -26,6 +26,7 @@
 
 #include "config.h"
 
+#include "galias.h"
 #include "glib.h"
 
 
@@ -662,7 +663,7 @@ g_queue_pop_nth_link (GQueue *queue,
   
   g_return_val_if_fail (queue != NULL, NULL);
 
-  if (n > queue->length)
+  if (n >= queue->length)
     return NULL;
   
   link = g_queue_peek_nth_link (queue, n);
@@ -729,7 +730,7 @@ gint
 g_queue_link_index (GQueue *queue,
 		    GList  *link_)
 {
-  g_return_val_if_fail (queue != NULL, 0);
+  g_return_val_if_fail (queue != NULL, -1);
 
   return g_list_position (queue->head, link_);
 }
